{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ScopedTypeVariables #-}
module Engine.Eval
( EvalResult(..)
, evalWithSeq
, evalVar
, compile
) where

import Control.Exception
import CoreLint
import CoreSyn
import CoreUtils (exprType)
import Data.List
import Engine.EngineHValue
import FastString
import GHC hiding (compileExpr, exprType)
import Outputable
import GHC.Prim
import HscMain (compileExpr)
import HscTypes
import Id
import MkCore
import Prelude hiding (catch)
import Util.GHC
import TysWiredIn (unitDataConId)
import UniqSupply
import Control.Applicative

data EvalResult 
  = EvalOk !Id !EngineHValue
  | EvalException !SomeException

data CompiledExpression = CE HValue !Type !Id
  
compile :: (GhcMonad m, MonadUnique m) => CoreExpr -> m CompiledExpression
compile expr = do
  let etipe = exprType expr
  letId <- mkSysLocalM (mkFastString "$irulan") etipe
  let rhs = mkCoreTup [Var unitDataConId, Var letId]
  let codeToRun = mkCoreLet (NonRec letId expr) $ 
                    mkWildCase (Var letId) etipe
                               (exprType rhs) [(DEFAULT, [], rhs)]
                                  
  
  --liftIO $ putStrLn $ showPpr codeToRun
  {-
  let mmsg = lintUnfolding noSrcLoc [] codeToRun
  case mmsg of
    Nothing -> return ()
    Just msg -> do liftIO $ putStrLn (showSDoc msg)
  -}
  {- builds:
   -  let letId = expr in 
   -    case letId of
   -      DEFAULT -> ((), letId)
   -}
  hval <- withSession $ \hsc -> liftIO $ compileExpr hsc noSrcSpan codeToRun 
  return $ CE hval etipe letId

evalWithSeq :: CompiledExpression -> IO EvalResult
evalWithSeq (CE hval etipe letId) = do
  case (unsafeCoerce# hval :: ((), HValue)) of
              ((),v) -> do
                ehv <- liftIO $ (mkEngineHValue v etipe)
                return $ EvalOk letId ehv
         `catch` \(e :: SomeException) -> do
            e' <- makeSafe e
            return (EvalException e')

makeSafeStr :: String -> IO String
makeSafeStr str = do
  (case str of
    ((!x):xs) -> do
      rest <- makeSafeStr xs
      rest `seq` return (x:rest)
    []     -> return "") `catch` (\(_ :: SomeException) -> return "<nested exception thrown>")
    
makeSafe :: SomeException -> IO SomeException
makeSafe e = (do
  () <- return (foldl' seq '!' str `seq` ())
  return e) `catch` (\(_ :: SomeException) -> do
    str' <- makeSafeStr str
    return (SomeException (ErrorCall str'))
    )
  where
    str = show e


evalVar :: (GhcMonad m) => Id -> m HValue
evalVar id = do
  withSession $ \hsc -> liftIO $ compileExpr hsc noSrcSpan (Var id)
