{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE BangPatterns #-}
module Engine.PlanStrategyTreePlan
( DepthLimitedRandomFork(..)
, DepthLimited(..)
) where

import Control.Applicative
import Control.Monad (guard)
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Engine.EngineHValue
import Engine.EngineM
import Engine.Experiment
import Engine.ExperimentResult
import Engine.RuntimePlan
import Format
import MonadUtils
import Plan
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Util.BiMap as BiMap
import qualified Util.UniqMap as UniqMap
import Util
import Util.BiMap (BiMap)
import Util.UniqMap (UniqMap)

data DepthLimitedRandomFork = DLRF { dlrfGetTreePlan    :: !TreePlan
                                   , dlrfDepthRemaining :: !Depth
                                   , dlrfFollowCase     :: !Bool
                                   }

instance PlanStrategy DepthLimitedRandomFork where
  select options = nextRandomR 0 (Seq.length options - 1)


  update (dlrfDepthRemaining -> x)
    | x <= 0 = return (Seq.empty, Nothing)

  update r@(dlrfGetTreePlan -> TPSplit { tpChildren = kids })
    = return $ (mkNext <$> kids, Nothing)
    where
      mkNext kid = r { dlrfGetTreePlan = kid
                     , dlrfDepthRemaining = nextDepth
                     }
      nextDepth = dlrfDepthRemaining r - 1

  update r@(dlrfGetTreePlan -> currentPlan@TPStep { tpExperiment = experiment })
    = updateFn currentPlan experiment (dlrfFollowCase r) mkNext
    where
      mkNext kid = r { dlrfGetTreePlan = kid
                     , dlrfDepthRemaining = nextDepth
                     }
      nextDepth = dlrfDepthRemaining r - 1

data DepthLimited = DL { dlGetTreePlan    :: !TreePlan
                       , dlDepthRemaining :: !Depth
                       , dlFollowCase     :: !Bool
                       }

instance PlanStrategy DepthLimited where
  select options = return 0

  update (dlDepthRemaining -> x)
    | x <= 0 = return (Seq.empty, Nothing)

  update r@(dlGetTreePlan -> TPSplit { tpChildren = kids })
    = return $ (mkNext <$> kids, Nothing)
    where
      mkNext kid = r { dlGetTreePlan = kid
                     , dlDepthRemaining = nextDepth
                     }
      nextDepth = dlDepthRemaining r - 1

  update r@(dlGetTreePlan -> currentPlan@TPStep { tpExperiment = experiment })
    = updateFn currentPlan experiment (dlFollowCase r) mkNext
    where
      mkNext kid = r { dlGetTreePlan = kid
                     , dlDepthRemaining = nextDepth
                     }
      nextDepth = dlDepthRemaining r - 1



updateFn :: TreePlan -> Experiment -> Bool -> (TreePlan -> a) -> EngineM (Seq.Seq a , Maybe (Experiment, RanExperiment))
updateFn currentPlan experiment followCase mkChild = do
  ranExperiment@(getExperimentResult -> result) <- runExperiment experiment

  let okLogic = case tpOnOk currentPlan of
                 Left p   -> return $ Seq.singleton p
                 Right mp -> case (guard followCase >> experimentRTValue ranExperiment) of
                               Nothing -> return $ Seq.empty
                               Just hv -> do
                                 mv <- if (UniqMap.null mp) 
                                         then return Nothing
                                         else getConstructor hv
                                 return $ maybe mempty (UniqMap.lookup mp) mv
  next <- case result of
            Ok          -> okLogic
            Bottoms bid -> return . fromMaybe (error "PlanStrategy updateFn onBottom") . Map.lookup bid . tpOnBottom $ currentPlan
            Property _  -> okLogic
            _           -> return $ Seq.empty
    
  return $ (mkChild <$> next, Just (experiment, ranExperiment))
