{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module FrontEnd.Source
(sourceAction
) where

import Bottom
import Control.Applicative
import Control.Monad
import Control.Monad.Identity
import Data.List
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq, (><), (<|), (|>))
import Data.Time.Clock
import Digraph
import Distribution.Simple.Setup
import DynFlags
import Engine
import Foreign
import Foreign.C.String
import Foreign.C.Types
import Format
import FrontEnd.Options
import FrontEnd.Stats
import GHC
import GHC.Paths
import HscTypes
import IdDB
import Linker
import ObjLink
import Outputable
import Plan as Plan
import PrelNames (unpackCStringName)
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import qualified Util.BiMap as BiMap
import System.FilePath
import System.IO
import Trace.Hpc.Mix
import Util
import Util.BiMap (BiMap)
import Util.GHC

foreign import ccall unsafe hs_hpc_module :: Ptr CChar -> Word32 -> Word32 -> Ptr Word64 -> IO ()

data PlanStrategyRep a where
  PSR :: (PlanStrategy a) => PlanStrategyRep a

sourceAction :: [String] -> [String] -> IrulanOptions -> IO ()
sourceAction ghcOpts targets options = do
  let args = a' ++ as
        where
          as = map (mkGeneralLocated "on the commandline") (concatMap words ghcOpts)
          a' = flagToList (const (mkGeneralLocated "on the commandline" "-fhpc") <$> outputStatistics options)

  (argsStatic, warnsStatic) <- parseStaticFlags args 

  runGhc (Just libdir) $ do
    (dflags', _, warns) <- flip parseDynamicFlags argsStatic =<< getSessionDynFlags 
    let dflags'' = dflags' { hscTarget = HscAsm 
                           , ghcLink   = LinkBinary
                           }
    primPackages <- setSessionDynFlags dflags''
    dflags <- getSessionDynFlags

    -- TODO change from this to one that says 'irulan blew up' or something
    defaultErrorHandler dflags $ do
      defaultCleanupHandler dflags $ do
          let userSpecifiedModules = map dropExtension targets
          let userSpecifiedExplicitSources = map dropExtension (explicitIdSource options)

          liftIO $ mapM_ (print . unLoc) (warns ++ warnsStatic)

          targets <- mapM (flip guessTarget Nothing) (nub $ userSpecifiedModules ++ userSpecifiedExplicitSources)
          setTargets targets

          result <- load LoadAllTargets

          liftIO $ initDynLinker dflags
          do
            dflags <- getSessionDynFlags
            liftIO $ linkPackages dflags primPackages
            -- For some reason, unpackCStringName# doesn't always get loaded.
            -- Force it to be so here.
            withSession $ \hsc -> (liftIO $ getHValue hsc unpackCStringName >> return ())
            return ()


          if (succeeded result) 
            then do
              mg <- depanal [] False
              let mods = flattenSCCs (topSortModuleGraph False mg Nothing)

              explicitIdentSourceMods <- mapM (flip lookupModule Nothing . mkModuleName) userSpecifiedExplicitSources

              {-
              (builtinSupply, constantsSupply) <- splitUniqSupply <$> (liftIO $ mkSplitUniqSupply 't')

              builtins <- do
                bs <- getBuiltins builtinSupply
                liftIO . newIORef $ {- if (intListHack options) then bs else -} []
              -}

              forM_ mods $ \modsum -> do
                printWarnings

                -- modInfo <- fromMaybe (error $ "Couldn't get module info: " ++ showPpr modsum) <$>
                                     -- (getModuleInfo (ms_mod modsum))
                modIface <- getModIface modsum
                
                let (modName :: String) = moduleNameString . ms_mod_name $ modsum

                linkModuleAndDependencies modsum modIface
                
                liftIO $ processHpc modName

                -- let availableIds = getAvailableIds modIface modsum modInfo
                
                when (modName `elem` userSpecifiedModules) $ do
                  liftIO $ putStrLn (modName ++ ":")

                  -- liftIO $ modifyIORef builtins (++ availableIds)
                  -- let functionsToTest  = filter (isNothing . isDataConWorkId_maybe) availableIds

                  -- liftIO . putStrLn $ "Testing: " ++ (unwords . map show $ functionsToTest)

                  -- supportFunctions <- (sortBy (comparing idNumArguments)) <$> liftIO (readIORef builtins )


                  theDB <- mkIdDBForModule (ms_mod modsum) (fromFlag $ sourceIdsFrom options) explicitIdentSourceMods

                  -- theDB <- addProviderIds supportFunctions <$> mkIdDBFor functionsToTest  
                  
                  errs <- testFunctions options theDB
                  liftIO $ processHpc modName

                  liftIO $ putStrLn "Results:"
                  liftIO . F.forM_ errs $ \(experiment, experimentResult) -> do
                    if (fromFlag $ useSimpleOutput options) 
                      then putStrLn ("  " ++ simpleOutput experiment (NewResult experimentResult Nothing))
                      else putStrLn ("  " ++ format experiment ++ " ==> " ++ format experimentResult)
                `gcatch` \(e :: SourceError) -> do
                           printExceptionAndWarnings e

            else do
              liftIO $ putStrLn "No!"

processHpc :: String -> IO ()
processHpc  modName = do
  modNamePtr <- newCString modName
  mTickboxesPtr <- lookupSymbol ("_hpc_tickboxes_" ++ modName ++ "_hpc")
  case mTickboxesPtr of
    Nothing -> return ()
    Just tickboxesPtr -> liftIO $ do
      (Mix _ _ infoHash _ (length -> infoTickCount)) <- readMix [".hpc"] (Left modName)
      let len = fromIntegral infoTickCount
      hs_hpc_module modNamePtr (len) (read . show $ infoHash) tickboxesPtr

linkModuleAndDependencies :: ModSummary -> ModIface -> Ghc ()
linkModuleAndDependencies modsum modiface = do
    dflags <- getSessionDynFlags
    liftIO $ linkPackages dflags (dep_pkgs . mi_deps $ modiface)
    let obj_file = ml_obj_file . ms_location $ modsum
    liftIO $ loadObj obj_file
    liftIO $ (resolveObjs >> return ())
    return ()


testFunctions :: IrulanOptions -> IdDB -> Ghc (Seq (Experiment, ExperimentResult))
testFunctions opts iddb = do
  let useCases = fromFlag $ useCaseStatement opts
  ExistsC PSR (Identity plan) <- return $ ((case (fromFlag $ runtimePlanType opts) of
                                              DepthBounded -> ExistsC PSR (Identity $ DL   (generateTreePlan iddb) (fromFlag $ maximumDepth opts) useCases)
                                              RandomTree   -> ExistsC PSR (Identity $ DLRF (generateTreePlan iddb) (fromFlag $ maximumDepth opts) useCases)) :: ExistsClass Identity PlanStrategyRep)

  engine <- buildEngine (fromJust . getLast $ cache opts)
                        (fromFlag $ maxExpressionTime opts) 
                        (fromFlag $ maxExpressionAllocs opts) plan

  unless (fromFlag $ outputExperimentTrace opts) (liftIO $ putStr " ")

  endTime <- case (fromFlag $ runtimePlanType opts) of
              RandomTree -> do timeNow <- liftIO $ getCurrentTime
                               let maxRuntime = (fromFlag $ maxRuntimePerFunction opts) * length (functionsToTest iddb)
                               return $ Just ((fromIntegral maxRuntime) `addUTCTime` timeNow)
              _          -> return Nothing

  withStatsOutput (flagToMaybe $ outputStatistics opts) $ \statsOutputter -> do
    res <- (engineLoop endTime engine (cycle ['▁'..'▏']) statsOutputter Seq.empty)
    unless ((fromFlag $ outputExperimentTrace opts) || (fromFlag $ quiet opts)) (liftIO $ putStr "\BS")
    return res
  where
    {- TODO: refactor this, it's a MESS! -}
    engineLoop :: (PlanStrategy a) => Maybe UTCTime -> Engine a -> [Char] -> StatsOutputter -> Seq (Experiment, ExperimentResult) -> Ghc (Seq (Experiment, ExperimentResult))
    engineLoop !endTime !engine (!c:cs) !statsOutputter !errors = do
      timeNow <- liftIO $ getCurrentTime

      if (isJust endTime) && (timeNow > fromJust endTime)
        then return errors
        else do
          let fmtFn
                | fromFlag $ useSimpleOutput opts = simpleOutput
                | otherwise                       = \exp ran -> let arr = case ran of
                                                                            NewResult _ _  -> " ==> "
                                                                            CachedResult _ -> " ##> "
                                                                  in format exp ++ arr ++ format (getExperimentResult ran)

          unless ((fromFlag $ outputExperimentTrace opts) || (fromFlag $ quiet opts)) $ do
              liftIO . putStr $ ['\BS', c]
              liftIO . hFlush $ stdout
          
          stepResult <- stepEngine engine

          case stepResult of
            NoWork -> return errors
            Work { experimentRun = experiment
                 , ranExperiment = ranExperiment
                 , nextEngine    = mNextEngine
                 , debugLog      = debugLog
                 } -> do
              when (fromFlag $ outputExperimentTrace opts) $ 
                (liftIO . putStrLn) (fmtFn experiment ranExperiment)
              when (fromFlag $ outputDebugInformation opts) $ 
                F.mapM_ (liftIO . putStrLn) debugLog
              liftIO $ outputStats statsOutputter experiment ranExperiment timeNow
              let nextErrors = case ranExperiment of
                    NewResult experimentResult@(Errors _) _       -> errors |> (experiment, experimentResult)
                    NewResult experimentResult@(Property False) _ -> errors |> (experiment, experimentResult)
                    _                                             -> errors
              case mNextEngine of
                Nothing        -> return nextErrors
                Just !nextEngine -> do
                  engineLoop endTime nextEngine cs statsOutputter $! nextErrors

simpleOutput :: Experiment -> RanExperiment -> String
simpleOutput exp' ranExp = simpleOutput' exp ++ arrow ++ simpleOutputResult res
  where
    (arrow, res')
      | NewResult res' _  <- ranExp = (" ==> ", res')
      | CachedResult res' <- ranExp = (" ##> ", res')

    (exp, bimap) = canonicaliseExperiment exp'
    res = remapExperimentResult (BiMap.forwardMap bimap) res'

    simpleOutput' exp 
      = unwords $ [simpleOutputRoot (experimentRoot exp)] ++ (catMaybes . F.toList $ simpleOutputArg <$> experimentArgs exp) 

    simpleOutputArg (ArgExperiment e)
      | Seq.null (experimentArgs e)  = Just $ simpleOutput' e
      | otherwise                    = Just $ "(" ++ simpleOutput' e ++ ")"
    simpleOutputArg (ArgBottom _ b)  = Just $ simpleBottom b
    simpleOutputArg (ArgType _)      = Nothing
    simpleOutputArg ArgUnconstrained = Just $ "*"

    simpleOutputRoot (IdRoot x) = showSDocUnqual . ppr $ x
    simpleOutputRoot (DestructRoot { destructCaseOf   
                                   , destructConstructor
                                   , destructVarIndex }) 
      = "case (" ++ (simpleOutput' destructCaseOf) ++ ") of " 
                ++ (showSDocUnqual . ppr $ destructConstructor) 
                ++ " -> " ++ show destructVarIndex

    simpleBottom b
      | b == initialBottomId         = "?"
      | otherwise                    = "?" ++ format b

    simpleOutputResult (Bottoms b) = simpleBottom b
    simpleOutputResult r           = format r
