{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell #-}
module Util.TH where

import BasicTypes
import Control.Applicative
import Control.Monad
import Control.Monad.RWS.Strict hiding (liftIO)
import Data.Map (Map)
import Data.Maybe
import GHC 
import HscTypes
import Id
import Language.Haskell.TH (TypeQ)
import qualified Data.Map as Map
import qualified Language.Haskell.TH as TH
import TcSplice
import Type
import TysPrim
import TysWiredIn
import Util.GHC
import VarEnv

typeFromTHTypeQ :: (GhcMonad m, MonadThings m) => TypeQ -> m Type
typeFromTHTypeQ tq = do
  thType <- liftIO $ TH.runQ tq
  typeFromTHType thType

typeFromTHType :: forall m . (GhcMonad m, MonadThings m) => TH.Type -> m Type
typeFromTHType thTy = fst <$> evalRWST (typeFromTHType' thTy) Map.empty emptyInScopeSet
  where
    typeFromTHType' :: TH.Type -> RWST (Map TH.Name TyVar) () InScopeSet m Type
    typeFromTHType' (TH.ForallT (map (\(TH.PlainTV n) -> n) -> names) [] thResTy) = do
      namesWithTyVars <- mapM mkTHTyVar names
      local (Map.fromList namesWithTyVars `Map.union`) $ 
        mkForAllTys (map snd namesWithTyVars) <$> typeFromTHType' thResTy
    typeFromTHType' (TH.VarT n)        = mkTyVarTy . fromJust <$> asks (Map.lookup n)
    typeFromTHType' (TH.TupleT 0)      = return . mkTyConTy $ unitTyCon
    typeFromTHType' (TH.TupleT n)      = return . mkTyConTy $ tupleTyCon Boxed n 
    typeFromTHType' (TH.ConT n)        = mkTyConTy <$> (lift $ tyconFromTHName n)
    typeFromTHType' (TH.ArrowT)        = return $ mkTyConTy funTyCon
    typeFromTHType' (TH.ListT)         = mkTyConTy <$> (lift $ tyconFromTHName ''[])
    typeFromTHType' (TH.AppT l r)      = liftM2 mkAppTy (typeFromTHType' l) 
                                                        (typeFromTHType' r)
    mkTHTyVar :: TH.Name -> RWST (Map TH.Name TyVar) () InScopeSet m (TH.Name, TyVar)
    mkTHTyVar thNam = do
      iss <- get
      let v' = uniqAway iss alphaTyVar 
      modify (flip extendInScopeSet v')
      return (thNam, v')

idFromTHName :: (GhcMonad m, MonadThings m) => TH.Name -> m Id
idFromTHName = thingFromTHName lookupId 

tyconFromTHName :: (GhcMonad m, MonadThings m) => TH.Name -> m TyCon
tyconFromTHName = thingFromTHName lookupTyCon

thingFromTHName :: (GhcMonad m, MonadThings m) => (Name -> m a) -> TH.Name -> m a
thingFromTHName f name = do
  hscEnv <- getSession
  (Just nam) <- liftIO $ initTcForLookup hscEnv (lookupThName_maybe name)
  f nam
